@@ -1,7 +1,7 @@
 /*****************************************************************************/
 /*  LibreDWG - free implementation of the DWG file format                    */
 /*                                                                           */
-/*  Copyright (C) 2009,2018-2019 Free Software Foundation, Inc.              */
+/*  Copyright (C) 2009,2018-2020 Free Software Foundation, Inc.              */
 /*                                                                           */
 /*  This library is free software, licensed under the terms of the GNU       */
 /*  General Public License as published by the Free Software Foundation,     */
@@ -31,7 +31,15 @@
 #endif
 // else we roll our own, Latin-1 only.
 
-#define DWG_LOGLEVEL DWG_LOGLEVEL_NONE
+#ifdef DWG_ABORT
+static unsigned int errors = 0;
+#  ifndef DWG_ABORT_LIMIT
+#    define DWG_ABORT_LIMIT 200
+#  endif
+#endif
+
+static unsigned int loglevel;
+#define DWG_LOGLEVEL loglevel
 #include "logging.h"
 #include "bits.h"
 
@@ -50,8 +58,9 @@ bit_advance_position (Bit_Chain *dat, long advance)
       // but allow pointing to the very end.
       if (dat->byte != dat->size - 1 || dat->bit != 0)
         {
-          LOG_ERROR ("buffer overflow at pos %lu, size %lu, advance by %ld",
-                     dat->byte, dat->size, advance)
+          loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+          LOG_ERROR ("%s buffer overflow at pos %lu, size %lu, advance by %ld",
+                     __FUNCTION__, dat->byte, dat->size, advance)
         }
       dat->byte = dat->size - 1;
       dat->bit = 0;
@@ -59,6 +68,7 @@ bit_advance_position (Bit_Chain *dat, long advance)
     }
   if ((long)dat->byte + (endpos / 8) < 0)
     {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("buffer underflow at pos %lu, size %lu, advance by %ld",
                  dat->byte, dat->size, advance)
       dat->byte = 0;
@@ -86,10 +96,24 @@ bit_set_position (Bit_Chain *dat, unsigned long bitpos)
   dat->bit = bitpos & 7;
   if (dat->byte > dat->size || (dat->byte == dat->size && dat->bit))
     {
-      LOG_ERROR ("buffer overflow at %lu, have %lu", dat->byte, dat->size)
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at %lu, have %lu", __FUNCTION__,
+                 dat->byte, dat->size)
     }
 }
 
+/* Set the chain so that dat->byte starts at 0 */
+void
+bit_reset_chain (Bit_Chain *dat)
+{
+  unsigned long pos = dat->byte;
+  dat->byte = 0;
+  if (pos < dat->size) // not already overflowed
+    dat->chain += pos;
+  if (dat->size > 0)
+    dat->size -= pos;
+}
+
 /** Read 1 bit.
  */
 BITCODE_B
@@ -100,7 +124,12 @@ bit_read_B (Bit_Chain *dat)
 
   if (dat->byte >= dat->size)
     {
-      LOG_ERROR ("buffer overflow at %lu", dat->byte)
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at %lu", __FUNCTION__, dat->byte)
+#ifdef DWG_ABORT
+      if (++errors > DWG_ABORT_LIMIT)
+        abort();
+#endif
       return 0;
     }
   byte = dat->chain[dat->byte];
@@ -136,7 +165,12 @@ bit_read_BB (Bit_Chain *dat)
 
   if (dat->byte >= dat->size)
     {
-      LOG_ERROR ("buffer overflow at %lu", dat->byte)
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at %lu", __FUNCTION__, dat->byte)
+#ifdef DWG_ABORT
+      if (++errors > DWG_ABORT_LIMIT)
+        abort();
+#endif
       return 0;
     }
   byte = dat->chain[dat->byte];
@@ -220,6 +254,7 @@ bit_write_3B (Bit_Chain *dat, unsigned char value)
 {
   if (value > 7)
     {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("Invalid bit_write_3B value %d > 7", value)
       return;
     }
@@ -236,28 +271,30 @@ bit_write_3B (Bit_Chain *dat, unsigned char value)
     }
 }
 
-/** Read 1 nibble.
+/** Read 4 bits.
  */
 BITCODE_4BITS
 bit_read_4BITS (Bit_Chain *dat)
 {
-  BITCODE_4BITS result = bit_read_RC (dat);
-  bit_advance_position (dat, -4);
-  // perhaps we have an issue here when the 4bit field is near the end of a
-  // bitstream?
-  // (since we initially advance 8bits and then later rewind 4bits)
-  return (result & 0xf0) >> 4;
+  // clang-format off
+  BITCODE_4BITS b =
+         bit_read_B (dat) << 3 |
+         bit_read_B (dat) << 2 |
+         bit_read_B (dat) << 1 |
+         bit_read_B (dat);
+  // clang-format on
+  return b;
 }
 
-/** Write 1 nibble.
+/** Write 4 bits.
  */
 void
 bit_write_4BITS (Bit_Chain *dat, unsigned char value)
 {
-  bit_write_B (dat, value & 1);
-  bit_write_B (dat, value & 2);
-  bit_write_B (dat, value & 4);
   bit_write_B (dat, value & 8);
+  bit_write_B (dat, value & 4);
+  bit_write_B (dat, value & 2);
+  bit_write_B (dat, value & 1);
 }
 
 /** Read 1 byte (raw char).
@@ -270,7 +307,12 @@ bit_read_RC (Bit_Chain *dat)
 
   if (dat->byte >= dat->size)
     {
-      LOG_ERROR ("buffer overflow at %lu", dat->byte)
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at %lu", __FUNCTION__, dat->byte)
+#ifdef DWG_ABORT
+      if (++errors > DWG_ABORT_LIMIT)
+        abort();
+#endif
       return 0;
     }
   byte = dat->chain[dat->byte];
@@ -284,6 +326,12 @@ bit_read_RC (Bit_Chain *dat)
           byte = dat->chain[dat->byte + 1];
           result |= byte >> (8 - dat->bit);
         }
+      else
+        {
+          loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+          LOG_ERROR ("%s buffer overflow at %lu", __FUNCTION__, dat->byte + 1)
+          return 0;
+        }
     }
 
   bit_advance_position (dat, 8);
@@ -519,6 +567,7 @@ bit_read_BL (Bit_Chain *dat)
     return 0;
   else /* if (two_bit_code == 3) */
     {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("bit_read_BL: unexpected 2-bit code: '11'")
       return 256;
     }
@@ -605,7 +654,7 @@ bit_write_BOT (Bit_Chain *dat, BITCODE_BS value)
     }
 }
 
-/** Read 1 bitlonglong (compacted uint64_t) for REQUIREDVERSIONS, picture_size.
+/** Read 1 bitlonglong (compacted uint64_t) for REQUIREDVERSIONS, preview_size.
  *  ODA doc bug. ODA say 1-3 bits until the first 0 bit. See 3BLL.
  *  The first 3 bits indicate the length l (see paragraph 2.1). Then
  *  l bytes follow, which represent the number (the least significant
@@ -720,6 +769,7 @@ bit_read_BD (Bit_Chain *dat)
     return 0.0;
   else /* if (two_bit_code == 3) */
     {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("bit_read_BD: unexpected 2-bit code: '11'")
       return bit_nan ();
     }
@@ -797,8 +847,13 @@ bit_read_MC (Bit_Chain *dat)
       result |= ((BITCODE_UMC)byte[i]) << j;
     }
 
-  LOG_ERROR ("bit_read_MC: error parsing modular char, i=%d,j=%d,result=0x%lx",
-             i, j, result)
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+  LOG_ERROR (
+      "bit_read_MC: error parsing modular char. i=%d, j=%d, result=0x%lx,\n"
+      " @%lu.@%u: [0x%x 0x%x 0x%x 0x%x 0x%x]",
+      i, j, result, dat->byte - 5, dat->bit, dat->chain[dat->byte - 5],
+      dat->chain[dat->byte - 4], dat->chain[dat->byte - 3],
+      dat->chain[dat->byte - 2], dat->chain[dat->byte - 1])
   return 0; /* error... */
 }
 
@@ -828,7 +883,7 @@ bit_write_MC (Bit_Chain *dat, BITCODE_MC val)
     if (byte[i] & 0x7f)
       break;
 
-  if (byte[i] & 0x40)
+  if (byte[i] & 0x40 && i > 0)
     i--;
   byte[i] &= 0x7f;
   if (negative)
@@ -837,17 +892,20 @@ bit_write_MC (Bit_Chain *dat, BITCODE_MC val)
     bit_write_RC (dat, byte[j]);
 }
 
-/** Read 1 modular char (max 5 bytes, unsigned).
+/** Read unsigned modular char (max 5 bytes, unsigned).
+    Can be quite large if there are many deleted handles.
  */
 BITCODE_UMC
 bit_read_UMC (Bit_Chain *dat)
 {
   int i, j;
-  unsigned char byte[5];
+  // eg handle FD485E65F
+  #define MAX_BYTE_UMC 6
+  unsigned char byte[MAX_BYTE_UMC];
   BITCODE_UMC result;
 
   result = 0;
-  for (i = 4, j = 0; i >= 0; i--, j += 7)
+  for (i = MAX_BYTE_UMC-1, j = 0; i >= 0; i--, j += 7)
     {
       byte[i] = bit_read_RC (dat);
       if (!(byte[i] & 0x80))
@@ -861,9 +919,14 @@ bit_read_UMC (Bit_Chain *dat)
       result |= ((BITCODE_UMC)byte[i]) << j;
     }
 
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
   LOG_ERROR (
       "bit_read_UMC: error parsing modular char, i=%d,j=%d,result=0x%lx", i, j,
       result)
+  LOG_HANDLE ("  @%lu.%u: [0x%x 0x%x 0x%x 0x%x 0x%x]\n", dat->byte - 5,
+            dat->bit, dat->chain[dat->byte - 5], dat->chain[dat->byte - 4],
+            dat->chain[dat->byte - 3], dat->chain[dat->byte - 2],
+            dat->chain[dat->byte - 1])
   return 0; /* error... */
 }
 
@@ -890,7 +953,7 @@ bit_write_UMC (Bit_Chain *dat, BITCODE_UMC val)
     if (byte[i] & 0x7f)
       break;
 
-  if (byte[i] & 0x40)
+  if (byte[i] & 0x40 && i > 0)
     i--;
   byte[i] &= 0x7f;
   for (j = 4; j >= i; j--)
@@ -919,6 +982,7 @@ bit_read_MS (Bit_Chain *dat)
         word[i] &= 0x7fff;
       result |= ((BITCODE_MS)word[i]) << j;
     }
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
   LOG_ERROR ("bit_read_MS: error parsing modular short, i=%d,j=%d", i, j)
   return 0; /* error... */
 }
@@ -1085,13 +1149,12 @@ bit_write_BT (Bit_Chain *dat, double value)
     }
 }
 
-/** Read handle-references. Returns error code: DWG_ERR_INVALIDHANDLE or 0 on
- * success
+/** Read handle-references. Returns error code: DWG_ERR_INVALIDHANDLE
+ *  or 0 on success.
  */
 int
 bit_read_H (Bit_Chain *restrict dat, Dwg_Handle *restrict handle)
 {
-  unsigned char *restrict val;
   int i;
 
   handle->code = bit_read_RC (dat);
@@ -1099,17 +1162,37 @@ bit_read_H (Bit_Chain *restrict dat, Dwg_Handle *restrict handle)
   handle->code = (handle->code & 0xf0) >> 4;
 
   handle->value = 0;
-  if (handle->size > 4 || handle->code > 14)
+  if (handle->size > sizeof(BITCODE_RC *) || handle->code > 14)
     {
-      LOG_WARN ("Invalid handle-reference, longer than 4 bytes: (%x.%d.%lX)",
-                handle->code, handle->size, handle->value)
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_WARN ("Invalid handle-reference, longer than 8 bytes: " FORMAT_H,
+                ARGS_H (*handle))
       return DWG_ERR_INVALIDHANDLE;
     }
 
   // TODO: little-endian only
-  val = (unsigned char *)&handle->value;
-  for (i = handle->size - 1; i >= 0; i--)
-    val[i] = bit_read_RC (dat);
+  // x86_64 gcc-9 miscompilation with val[i]: (%rbx) being dat+1
+#if defined(__GNUC__) && (__GNUC__ == 9) && (__GNUC_MINOR__ <= 2) \
+  && (SIZEOF_SIZE_T == 8)                                         \
+  && (defined(__x86_64__) || defined(__x86_64) || defined(_M_AMD64) || defined(_M_X64))
+  {
+#warning x86_64 gcc-9.2 codegen is seriously broken! better use an older version
+    BITCODE_RC *restrict val;
+    val = (BITCODE_RC *)&(handle->value);
+    for (i = handle->size - 1; i >= 0; i--)
+      {
+        BITCODE_RC c = bit_read_RC (dat);
+        val[i] = c;
+      }
+  }
+#else
+  {
+    BITCODE_RC *restrict val;
+    val = (BITCODE_RC *)&(handle->value);
+    for (i = handle->size - 1; i >= 0; i--)
+      val[i] = bit_read_RC (dat);
+  }
+#endif
 
   return 0;
 }
@@ -1130,19 +1213,19 @@ bit_write_H (Bit_Chain *restrict dat, Dwg_Handle *restrict handle)
     }
   if (handle->value == 0)
     {
-      bit_write_RC (dat, (handle->code << 4));
+      bit_write_RC (dat, handle->code << 4);
       return;
     }
 
-  // TODO: little-endian only
+  // TODO: little-endian only. support sizes <= 8, not just 4
+  memset (&val, 0, sizeof(val));
   val = (unsigned char *)&handle->value;
-  for (i = 3; i >= 0; i--)
+  for (i = sizeof(val) - 1; i >= 0; i--)
     if (val[i])
       break;
 
   size = handle->code << 4;
   size |= i + 1;
-
   bit_write_RC (dat, size);
 
   for (; i >= 0; i--)
@@ -1156,16 +1239,15 @@ uint16_t
 bit_read_CRC (Bit_Chain *dat)
 {
   uint16_t result;
-  long unsigned int start_address;
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
 
   if (dat->bit > 0)
     {
       dat->byte++;
       dat->bit = 0;
     }
-  start_address = dat->byte;
   result = bit_read_RS (dat);
-  LOG_TRACE ("read CRC at %lx: 0x%x\n", start_address, result)
+  LOG_TRACE ("read CRC at %lu: %04X\n", dat->byte, result)
 
   return result;
 }
@@ -1177,6 +1259,8 @@ bit_check_CRC (Bit_Chain *dat, long unsigned int start_address, uint16_t seed)
 {
   uint16_t calculated;
   uint16_t read;
+  long size;
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
 
   if (dat->bit > 0)
     {
@@ -1184,13 +1268,29 @@ bit_check_CRC (Bit_Chain *dat, long unsigned int start_address, uint16_t seed)
       dat->bit = 0;
     }
 
-  calculated = bit_calc_CRC (seed, &(dat->chain[start_address]),
-                             dat->byte - start_address);
+  if (start_address > dat->byte || dat->byte >= dat->size)
+    {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at pos %lu-%lu, size %lu", __FUNCTION__,
+                 start_address, dat->byte, dat->size)
+      return 0;
+    }
+  size = dat->byte - start_address;
+  calculated = bit_calc_CRC (seed, &dat->chain[start_address], size);
   read = bit_read_RS (dat);
-  LOG_TRACE ("check CRC at %lx: 0x%x <=> 0x%x\n", start_address, calculated,
-             read)
-
-  return (calculated == read);
+  LOG_TRACE ("crc: %04X [RSx]\n", read);
+  if (calculated == read)
+    {
+      LOG_HANDLE (" check_CRC %lu-%lu = %ld: %04X == %04X\n", start_address,
+                  dat->byte - 2, size, calculated, read)
+      return 1;
+    }
+  else
+    {
+      LOG_WARN ("check_CRC mismatch %lu-%lu = %ld: %04X <=> %04X\n",
+                start_address, dat->byte - 2, size, calculated, read)
+      return 0;
+    }
 }
 
 /** Create and write old 16bit CRC.
@@ -1199,24 +1299,76 @@ uint16_t
 bit_write_CRC (Bit_Chain *dat, long unsigned int start_address, uint16_t seed)
 {
   uint16_t crc;
+  long size;
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
 
   while (dat->bit > 0)
     bit_write_B (dat, 0);
 
-  crc = bit_calc_CRC (seed, &(dat->chain[start_address]),
-                      dat->byte - start_address);
+  if (start_address > dat->byte || dat->byte >= dat->size)
+    {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at pos %lu-%lu, size %lu", __FUNCTION__,
+                 start_address, dat->byte, dat->size)
+      return 0;
+    }
+  size = dat->byte - start_address;
+  crc = bit_calc_CRC (seed, &dat->chain[start_address], size);
 
+  LOG_TRACE ("write CRC %04X from %lu-%lu = %ld\n", crc, start_address,
+             dat->byte, size);
   bit_write_RS (dat, crc);
-  LOG_TRACE ("write CRC at %lx: %04X\n", start_address, crc)
-  return (crc);
+  return crc;
+}
+
+uint16_t
+bit_write_CRC_LE (Bit_Chain *dat, long unsigned int start_address, uint16_t seed)
+{
+  uint16_t crc;
+  long size;
+
+  while (dat->bit > 0)
+    bit_write_B (dat, 0);
+
+  if (start_address > dat->byte || dat->byte >= dat->size)
+    {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at pos %lu-%lu, size %lu", __FUNCTION__,
+                 start_address, dat->byte, dat->size)
+      return 0;
+    }
+  size = dat->byte - start_address;
+  crc = bit_calc_CRC (seed, &dat->chain[start_address], size);
+
+  loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+  LOG_TRACE ("write CRC %04X from %lu-%lu = %ld\n", crc, start_address,
+             dat->byte, size);
+  bit_write_RS_LE (dat, crc);
+  return crc;
 }
 
 void
-bit_read_fixed (Bit_Chain *restrict dat, BITCODE_RC *restrict dest, int length)
+bit_read_fixed (Bit_Chain *restrict dat, BITCODE_RC *restrict dest,
+                unsigned int length)
 {
-  for (int i = 0; i < length; i++)
+  if (dat->byte + length >= dat->size)
     {
-      dest[i] = bit_read_RC (dat);
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at pos %lu, size %lu", __FUNCTION__,
+                 dat->byte, dat->size)
+      return;
+    }
+  if (dat->bit == 0)
+    {
+      memcpy (dest, &dat->chain[dat->byte], length);
+      dat->byte += length;
+    }
+  else
+    {
+      for (unsigned int i = 0; i < length; i++)
+        {
+          dest[i] = bit_read_RC (dat);
+        }
     }
 }
 
@@ -1224,10 +1376,11 @@ bit_read_fixed (Bit_Chain *restrict dat, BITCODE_RC *restrict dest, int length)
  *  After usage, the allocated memory must be properly freed.
  *  preR11
  */
+ATTRIBUTE_MALLOC
 BITCODE_TF
-bit_read_TF (Bit_Chain *restrict dat, int length)
+bit_read_TF (Bit_Chain *restrict dat, unsigned int length)
 {
-  BITCODE_RC *chain = malloc (length + 1);
+  BITCODE_RC *chain = calloc (length + 1, 1);
 
   bit_read_fixed (dat, chain, length);
   chain[length] = '\0';
@@ -1235,14 +1388,21 @@ bit_read_TF (Bit_Chain *restrict dat, int length)
   return (BITCODE_TF)chain;
 }
 
-/** Write fixed text.
+/** Write fixed text. Without ending \0
  */
 void
-bit_write_TF (Bit_Chain *restrict dat, BITCODE_TF restrict chain, int length)
+bit_write_TF (Bit_Chain *restrict dat, BITCODE_TF restrict chain, unsigned int length)
 {
-  int i;
-  for (i = 0; i < length; i++)
-    bit_write_RC (dat, (BITCODE_RC)chain[i]);
+  if (dat->bit == 0 && dat->byte + length < dat->size)
+    {
+      memcpy (&dat->chain[dat->byte], chain, length);
+      dat->byte += length;
+    }
+  else
+    {
+      for (unsigned int i = 0; i < length; i++)
+        bit_write_RC (dat, (BITCODE_RC)chain[i]);
+    }
 }
 
 /** Read simple text. After usage, the allocated memory must be properly freed.
@@ -1255,7 +1415,13 @@ bit_read_TV (Bit_Chain *restrict dat)
   unsigned char *chain;
 
   length = bit_read_BS (dat);
-  // if (length > AVAIL_BITS()) return DWG_ERR_VALUEOUTOFBOUNDS;
+  if (dat->byte + length >= dat->size)
+    {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at %lu, length %u", __FUNCTION__,
+                 dat->byte, length)
+      return NULL;
+    }
   chain = (unsigned char *)malloc (length + 1);
   for (i = 0; i < length; i++)
     {
@@ -1274,7 +1440,7 @@ bit_read_TV (Bit_Chain *restrict dat)
 /** Write simple text.
  */
 void
-bit_write_TV (Bit_Chain *restrict dat, char *restrict chain)
+bit_write_TV (Bit_Chain *restrict dat, BITCODE_TV restrict chain)
 {
   int i;
   int length;
@@ -1296,6 +1462,13 @@ bit_read_TU (Bit_Chain *restrict dat)
   BITCODE_TU chain;
 
   length = bit_read_BS (dat);
+  if (dat->byte + (length * 2) >= dat->size)
+    {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("%s buffer overflow at %lu, length %u", __FUNCTION__,
+                 dat->byte, length)
+      return NULL;
+    }
   chain = (BITCODE_TU)malloc ((length + 1) * 2);
   for (i = 0; i < length; i++)
     {
@@ -1335,6 +1508,24 @@ bit_write_TU (Bit_Chain *restrict dat, BITCODE_TU restrict chain)
   bit_write_RS (dat, 0); //?? unsure about that
 }
 
+BITCODE_T
+bit_read_T (Bit_Chain *restrict dat)
+{
+  if (dat->version >= R_2007)
+    return (BITCODE_T)bit_read_TU (dat);
+  else
+    return (BITCODE_T)bit_read_TV (dat);
+}
+
+void
+bit_write_T (Bit_Chain *restrict dat, BITCODE_T restrict chain)
+{
+  if (dat->version >= R_2007)
+    return bit_write_TU (dat, (BITCODE_TU)chain);
+  else
+    return bit_write_TV (dat, chain);
+}
+
 /* converts UCS-2 to UTF-8 */
 char *
 bit_convert_TU (BITCODE_TU restrict wstr)
@@ -1346,6 +1537,25 @@ bit_convert_TU (BITCODE_TU restrict wstr)
 
   if (!wstr)
     return NULL;
+#ifdef HAVE_ALIGNED_ACCESS_REQUIRED
+  /* access bytewise. cannot copy to heap yet, because we have no length */
+  if ((uintptr_t)wstr % SIZEOF_SIZE_T)
+    {
+      unsigned char *b = (unsigned char*)wstr;
+      c = (b[0] << 8) + b[1];
+      while (c)
+        {
+          len++;
+          if (c >= 256)
+            len++;
+          if (c >= 0x800)
+            len++;
+          b += 2;
+          c = (b[0] << 8) + b[1];
+        }
+    }
+  else
+#endif
   while ((c = *tmp++))
     {
       len++;
@@ -1353,38 +1563,74 @@ bit_convert_TU (BITCODE_TU restrict wstr)
         len++;
       if (c >= 0x800)
         len++;
+#if 0
+        loglevel = 5;
+        LOG_INSANE ("U+%04X ", c);
+#endif
     }
   str = malloc (len + 1);
   i = 0;
+#ifdef HAVE_ALIGNED_ACCESS_REQUIRED
+  if ((uintptr_t)wstr % SIZEOF_SIZE_T)
+    {
+      unsigned char *b = (unsigned char*)wstr;
+      c = (b[0] << 8) + b[1];
+      while (c)
+        {
+          if (c < 256)
+            {
+              str[i++] = c & 0xFF;
+            }
+          else if (c < 0x800)
+            {
+              str[i++] = (c >> 6) | 0xC0;
+              str[i++] = (c & 0x3F) | 0x80;
+            }
+          else /* if (c < 0x10000) */
+            {
+              str[i++] = (c >> 12) | 0xE0;
+              str[i++] = ((c >> 6) & 0x3F) | 0x80;
+              str[i++] = (c & 0x3F) | 0x80;
+            }
+
+          b += 2;
+          c = (b[0] << 8) + b[1];
+        }
+    }
+  else
+#endif
   while ((c = *wstr++))
     {
       if (c < 256)
         {
-          str[i++] = c & 0xff;
+          str[i++] = c & 0xFF;
         }
       else if (c < 0x800)
         {
+          str[i++] = (c >> 6) | 0xC0;
+          str[i++] = (c & 0x3F) | 0x80;
+        }
+      else /* if (c < 0x10000) */
+        { /* windows ucs-2 has no D800-DC00 surrogate pairs. go straight up */
           /*if (i+3 > len) {
             str = realloc(str, i+3);
             len = i+2;
           }*/
-          str[i + 1] = (c & 0x3f) | 0x80;
-          str[i] = (c >> 6) | 0xc0;
-          i += 2;
+          str[i++] = (c >> 12) | 0xE0;
+          str[i++] = ((c >> 6) & 0x3F) | 0x80;
+          str[i++] = (c & 0x3F) | 0x80;
+        }
+      /*
+      else if (c < 0x110000)
+        {
+          str[i++] = (c >> 18) | 0xF0;
+          str[i++] = ((c >> 12) & 0x3F) | 0x80;
+          str[i++] = ((c >> 6) & 0x3F) | 0x80;
+          str[i++] = (c & 0x3F) | 0x80;
         }
       else
-        { /* windows ucs-2 has no D800-DC00 surrogate pairs. go straight up */
-          /*if (i+3 > len) {
-            str = realloc(str, i+4);
-            len = i+3;
-          }*/
-          str[i + 2] = (c & 0x3f) | 0x80;
-          c >>= 6;
-          str[i + 1] = (c & 0x3f) | 0x80;
-          c >>= 6;
-          str[i] = c | 0xe0;
-          i += 3;
-        }
+        fprintf (stderr, "ERROR: overlarge unicode codepoint U+%0X", c);
+     */
     }
   str[i] = '\0';
   return str;
@@ -1400,8 +1646,9 @@ bit_utf8_to_TU (char *restrict str)
   unsigned char c;
 
   wstr = malloc (2 * (len + 1));
-  while ((c = *str++))
+  while (len >= 0 && (c = *str++))
     {
+      len--;
       if (c < 128)
         {
           wstr[i++] = c;
@@ -1409,24 +1656,31 @@ bit_utf8_to_TU (char *restrict str)
       else if ((c & 0xe0) == 0xc0)
         {
           /* ignore invalid utf8 for now */
-          wstr[i++] = ((c & 0x1f) << 6) | (str[1] & 0x3f);
+          if (len >= 1)
+            wstr[i++] = ((c & 0x1f) << 6) | (str[1] & 0x3f);
+          len--;
           str++;
         }
       else if ((c & 0xf0) == 0xe0)
         {
-          /* ignore invalid utf8 for now */
-          /*
-          if (str[1] < 0x80 || str[1] > 0xBF ||
-              str[2] < 0x80 || str[2] > 0xBF) {
-            LOG_ERROR("utf-8: BAD_CONTINUATION_BYTE %s", str);
-          }
-          if (c == 0xe0 && str[1] < 0xa0) {
-            LOG_ERROR("utf-8: NON_SHORTEST %s", str);
-          } */
-          wstr[i++]
+          /* ignore invalid utf8? */
+          if (len >= 2 &&
+              ((unsigned char)str[1] < 0x80 || (unsigned char)str[1] > 0xBF
+               || (unsigned char)str[2] < 0x80 || (unsigned char)str[2] > 0xBF))
+            {
+              LOG_WARN ("utf-8: BAD_CONTINUATION_BYTE %s", str);
+            }
+          if (len >= 1 && c == 0xe0 && (unsigned char)str[1] < 0xa0)
+            {
+              LOG_WARN ("utf-8: NON_SHORTEST %s", str);
+            }
+          if (len >= 2)
+            wstr[i++]
               = ((c & 0x0f) << 12) | ((str[1] & 0x3f) << 6) | (str[2] & 0x3f);
           str++;
           str++;
+          len--;
+          len--;
         }
       /* everything above 0xf0 exceeds ucs-2, 4-6 byte seqs */
     }
@@ -1504,16 +1758,42 @@ bit_write_TIMEBLL (Bit_Chain *dat, BITCODE_TIMEBLL date)
     }
 }
 
+BITCODE_TIMERLL
+bit_read_TIMERLL (Bit_Chain *dat)
+{
+  BITCODE_TIMERLL date;
+  BITCODE_BD ms;
+  date.days = bit_read_RL (dat);
+  date.ms = bit_read_RL (dat);
+  ms = (double)date.ms;
+  while (ms > 1.0)
+    ms /= 10.0;
+  date.value = date.days + ms; // just for display, not calculations
+  return date;
+}
+
+/** Write 2x time RL.
+ *  Ignores the double value.
+ */
+void
+bit_write_TIMERLL (Bit_Chain *dat, BITCODE_TIMERLL date)
+{
+  bit_write_RL (dat, date.days);
+  bit_write_RL (dat, date.ms);
+}
+
 /** Read color
  */
 void
 bit_read_CMC (Bit_Chain *restrict dat, Dwg_Color *restrict color)
 {
+  memset (color, 0, sizeof (Dwg_Color));
   color->index = bit_read_BS (dat);
   if (dat->version >= R_2004)
     {
       color->rgb = bit_read_BL (dat);
       color->flag = bit_read_RC (dat);
+      // wide?
       color->name = (color->flag & 1) ? (char *)bit_read_TV (dat) : NULL;
       color->book_name = (color->flag & 2) ? (char *)bit_read_TV (dat) : NULL;
     }
@@ -1529,6 +1809,7 @@ bit_write_CMC (Bit_Chain *restrict dat, Dwg_Color *restrict color)
     {
       bit_write_BL (dat, color->rgb);
       bit_write_RC (dat, color->flag);
+      // wide?
       if (color->flag & 1)
         bit_write_TV (dat, color->name);
       if (color->flag & 2)
@@ -1537,22 +1818,24 @@ bit_write_CMC (Bit_Chain *restrict dat, Dwg_Color *restrict color)
 }
 
 /** Read entity color (2004+) (truecolor rgb and alpha support)
- *  Does also references, DBCOLOR lookups (not yet, needs hdl_dat stream)
+ *  Does also references, DBCOLOR lookups.
  */
 void
-bit_read_ENC (Bit_Chain *restrict dat, Dwg_Color *restrict color)
+bit_read_ENC (Bit_Chain *dat, Bit_Chain *hdl_dat, Bit_Chain *str_dat,
+              Dwg_Color *restrict color)
 {
   color->index = bit_read_BS (dat);
   if (dat->version >= R_2004)
     {
-      uint16_t flag = color->index >> 8;
+      uint16_t flag = (((uint32_t)color->index) >> 8) & 0xff;
       color->index &= 0x1ff;
       if (flag & 0x80)
         color->rgb = bit_read_BL (dat); // ODA bug, documented as BS
       if (flag & 0x40)
         {
-          color->handle = calloc (1, sizeof (color->handle));
-          bit_read_H (dat, color->handle); // => DBCOLOR
+          color->handle = calloc (1, sizeof (Dwg_Object_Ref));
+          bit_read_H (hdl_dat, &(color->handle->handleref)); // => DBCOLOR
+          // else defer to dwg_decode_common_entity_handle_data ()
         }
       if (flag & 0x20)
         {
@@ -1560,32 +1843,31 @@ bit_read_ENC (Bit_Chain *restrict dat, Dwg_Color *restrict color)
           color->alpha_type = alpha & 0xff; // 0, 1 or 3
           color->alpha = alpha >> 8;
         }
-      color->flag = flag;
+      color->flag = (uint16_t)flag;
     }
 }
 
 /** Write entity color (2004+)
  */
 void
-bit_write_ENC (Bit_Chain *restrict dat, Dwg_Color *restrict color)
+bit_write_ENC (Bit_Chain *dat, Bit_Chain *hdl_dat, Bit_Chain *str_dat,
+               Dwg_Color *restrict color)
 {
   bit_write_BS (dat, (color->index & 0x1ff) | (color->flag << 8));
   if (dat->version >= R_2004)
     {
       uint16_t flag = color->flag;
-      // if (flag & 0x40) {
-      //  color.handle = dwg_decode_handleref(dat, obj, dwg); //hdl_dat!
-      //}
       if (flag & 0x20)
-        {
-          bit_write_BL (dat, color->alpha);
-        }
+        bit_write_BL (dat, color->alpha);
       if (!(flag & 0x40) && (flag & 0x80))
         bit_write_BL (dat, color->rgb);
+      // ?? wide?
       if ((flag & 0x41) == 0x41)
-        bit_write_TV (dat, color->name); // str_dat
+        bit_write_T (str_dat, color->name);
       if ((flag & 0x42) == 0x42)
-        bit_write_TV (dat, color->book_name); // str_dat
+        bit_write_T (str_dat, color->book_name);
+      if (flag & 0x40)
+        bit_write_H (hdl_dat, &(color->handle->handleref)); // => DBCOLOR
     }
 }
 
@@ -1597,7 +1879,9 @@ bit_search_sentinel (Bit_Chain *dat, unsigned char sentinel[16])
 {
   long unsigned int i, j;
 
-  for (i = 0; i < dat->size; i++)
+  if (dat->size < 16) // too short
+    return 0;
+  for (i = 0; i < dat->size - 16; i++)
     {
       for (j = 0; j < 16; j++)
         {
@@ -1642,6 +1926,7 @@ bit_chain_alloc (Bit_Chain *dat)
     {
       dat->chain
           = (unsigned char *)realloc (dat->chain, dat->size + CHAIN_BLOCK);
+      memset (&dat->chain[dat->size], 0, CHAIN_BLOCK);
       dat->size += CHAIN_BLOCK;
     }
 }
@@ -1681,10 +1966,60 @@ bit_write_bits (Bit_Chain *restrict dat, const char *restrict bits)
   char *p = (char *)bits;
   for (; *p; p++)
     {
-      bit_write_B (dat, *p != '0');
+      if (*p == '0' || *p == '1')
+        bit_write_B (dat, *p != '0');
+      else
+        {
+          fprintf (stderr, "Invalid binary input %s\n", p);
+          return;
+        }
     }
 }
 
+// accept a string of hex bytes with optional whitespace
+long
+bit_write_hexbits (Bit_Chain *restrict dat, const char *restrict bytes)
+{
+  char *p = (char *)bytes;
+  long len = 0;
+  unsigned char b = '\0';
+
+  for (; *p; p++)
+    {
+      if (*p != ' ' && *p != '\n')
+        {
+          len++;
+          if (*p >= 'a' && *p <= 'f')
+            {
+              if (len % 2)
+                b = (*p + 10 - 'a') << 4;
+              else
+                bit_write_RC (dat, b + *p + 10 - 'a');
+            }
+          else if (*p >= 'A' && *p <= 'F')
+            {
+              if (len % 2)
+                b = (*p + 10 - 'A') << 4;
+              else
+                bit_write_RC (dat, b + *p + 10 - 'A');
+            }
+          else if (*p >= '0' && *p <= '9')
+            {
+              if (len % 2)
+                b = (*p - '0') << 4;
+              else
+                bit_write_RC (dat, b + *p - '0');
+            }
+          else
+            {
+              fprintf (stderr, "Invalid hex input %s\n", p);
+              return 0;
+            }
+        }
+    }
+  return len;
+}
+
 void
 bit_print_bits (unsigned char *bits, long unsigned int size)
 {
