@@ -1,7 +1,7 @@
 /*****************************************************************************/
 /*  LibreDWG - free implementation of the DWG file format                    */
 /*                                                                           */
-/*  Copyright (C) 2009-2019 Free Software Foundation, Inc.                   */
+/*  Copyright (C) 2009-2020 Free Software Foundation, Inc.                   */
 /*                                                                           */
 /*  This library is free software, licensed under the terms of the GNU       */
 /*  General Public License as published by the Free Software Foundation,     */
@@ -86,6 +86,21 @@
 #  define __has_feature(x) 0
 #endif
 
+/* */
+#if defined(__AFL_COMPILER) && defined(__clang__)
+#  define AFL_GCC_TOOBIG __attribute__((optnone))
+#  define AFL_GCC_POP
+#elif defined(__AFL_COMPILER) && defined(__GNUC__)
+#  define AFL_GCC_TOOBIG \
+    _Pragma ("GCC push_options") \
+    _Pragma ("GCC optimize (\"-fno-var-tracking-assignments\")")
+#  define AFL_GCC_POP \
+    _Pragma ("GCC pop_options")
+#else
+#  define AFL_GCC_TOOBIG
+#  define AFL_GCC_POP
+#endif
+
 /* The __nonnull function attribute marks pointer arguments which
    must not be NULL.  */
 #if (defined(__GNUC__) && ((__GNUC__ * 100) + __GNUC_MINOR__) >= 303)
@@ -103,6 +118,12 @@
 #  define ATTRIBUTE_MALLOC
 #endif
 
+#ifdef HAVE_FUNC_ATTRIBUTE_RETURNS_NONNULL
+#  define RETURNS_NONNULL __attribute__ ((returns_nonnull))
+#else
+#  define RETURNS_NONNULL
+#endif
+
 #ifndef EXPORT
 #  if defined(_WIN32) && defined(ENABLE_SHARED)
 #    ifdef DLL_EXPORT
