@@ -1,7 +1,7 @@
 /*****************************************************************************/
 /*  LibreDWG - free implementation of the DWG file format                    */
 /*                                                                           */
-/*  Copyright (C) 2018-2019 Free Software Foundation, Inc.                   */
+/*  Copyright (C) 2018-2020 Free Software Foundation, Inc.                   */
 /*                                                                           */
 /*  This library is free software, licensed under the terms of the GNU       */
 /*  General Public License as published by the Free Software Foundation,     */
@@ -17,11 +17,13 @@
  */
 
 #include "config.h"
-#include <stdio.h>
-#include <stdlib.h>
 #ifdef __STDC_ALLOC_LIB__
 #  define __STDC_WANT_LIB_EXT2__ 1 /* for strdup */
+#else
+#  define _USE_BSD 1
 #endif
+#include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <assert.h>
 
@@ -57,6 +59,11 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
     ptr = NULL;                                                               \
   }
 
+#undef SINCE
+#define SINCE(v)                                                              \
+  cur_ver = v;                                                                \
+  if (dat->from_version >= v)
+
 #define VALUE(value, type, dxf)
 #define VALUE_RC(value, dxf) VALUE (value, RC, dxf)
 #define VALUE_RS(value, dxf) VALUE (value, RS, dxf)
@@ -81,11 +88,17 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 #define FIELD_HANDLE(nam, code, dxf) VALUE_HANDLE (_obj->nam, nam, code, dxf)
 #define SUB_FIELD_HANDLE(o, nam, code, dxf)                                   \
   VALUE_HANDLE (_obj->o.nam, nam, code, dxf)
+// compare to dwg_decode_handleref_with_code: not all refs are stored in the
+// object_ref vector, like relative ptrs and NULL.
+// But indxf skip the NULL HDL, it is global and shared there.
+// obj is the relative base object here and there.
 #define VALUE_HANDLE(ref, nam, _code, dxf)                                    \
-  if (ref)                                                                    \
+  if (ref                                                                     \
+      && !(dat->opts & DWG_OPTS_INDXF && ref->handleref.size == 0             \
+           && ref->absolute_ref == 0 && !ref->obj)                            \
+      && !(ref->handleref.size || (obj && ref->handleref.code > 5)))          \
     {                                                                         \
-      if (!(ref->handleref.size || (obj && ref->handleref.code > 5)))         \
-        free (ref);                                                           \
+      free (ref);                                                             \
       ref = NULL;                                                             \
     } /* else freed globally */
 #define FIELD_DATAHANDLE(name, code, dxf) FIELD_HANDLE (name, code, dxf)
@@ -105,11 +118,7 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 #define FIELD_RLL(name, dxf) FIELD (name, RLL)
 #define FIELD_MC(name, dxf) FIELD (name, MC)
 #define FIELD_MS(name, dxf) FIELD (name, MS)
-#define FIELD_TV(name, dxf)                                                   \
-  if (FIELD_VALUE (name))                                                     \
-    {                                                                         \
-      FREE_IF (FIELD_VALUE (name));                                           \
-    }
+#define FIELD_TV(name, dxf) FREE_IF (FIELD_VALUE (name))
 #define VALUE_TV(value, dxf) FREE_IF (value)
 #define VALUE_TF(value, dxf) FREE_IF (value)
 #define VALUE_TFF(value, dxf)
@@ -119,6 +128,7 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
   {                                                                           \
   }
 #define FIELD_T(name, dxf) FIELD_TV (name, dxf)
+#define FIELD_BINARY(name, len, dxf) FIELD_TV (name, dxf)
 #define FIELD_BT(name, dxf) FIELD (name, BT);
 #define FIELD_4BITS(name, dxf)                                                \
   {                                                                           \
@@ -157,13 +167,15 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
   {                                                                           \
   }
 #define FIELD_TIMEBLL(name, dxf)
+#define FIELD_TIMERLL(name, dxf)
+// indxf does not assign names yet
 #define FIELD_CMC(color, dxf1, dxf2)                                          \
-  {                                                                           \
-    FIELD_TV (color.name, 0);                                                 \
-    FIELD_TV (color.book_name, 0);                                            \
+  if (!(dat->opts & DWG_OPTS_INDXF)) {                                        \
+    FIELD_T (color.name, 0);                                                  \
+    FIELD_T (color.book_name, 0);                                             \
   }
 #define SUB_FIELD_CMC(o, color, dxf1, dxf2)                                   \
-  {                                                                           \
+  if (!(dat->opts & DWG_OPTS_INDXF)) {                                        \
     VALUE_TV (_obj->o.color.name, 0);                                         \
     VALUE_TV (_obj->o.color.book_name, 0);                                    \
   }
@@ -171,13 +183,13 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 // FIELD_VECTOR_N(name, type, size):
 // reads data of the type indicated by 'type' 'size' times and stores
 // it all in the vector called 'name'.
-#define FIELD_VECTOR_N(name, type, size, dxf)                                 \
-  if (size && _obj->name)                                                     \
+#define FIELD_VECTOR_N(nam, type, size, dxf)                                  \
+  if ((size) && _obj->nam)                                                    \
     {                                                                         \
-      for (vcount = 0; vcount < (BITCODE_BL)size; vcount++)                   \
-        FIELD_##type (name[vcount], dxf);                                     \
+      for (vcount = 0; vcount < (BITCODE_BL) (size); vcount++)                \
+        FIELD_##type (nam[vcount], dxf);                                      \
     }                                                                         \
-  FIELD_TV (name, dxf);
+  FIELD_TV (nam, dxf);
 #define FIELD_VECTOR_T(name, size, dxf)                                       \
   FIELD_VECTOR_N (name, TV, _obj->size, dxf)
 #define FIELD_VECTOR(name, type, size, dxf)                                   \
@@ -243,8 +255,7 @@ static Bit_Chain pdat = { NULL, 0, 0, 0, 0, 0 };
 
 #define END_REPEAT(field) FIELD_TV (field, 0)
 
-#define COMMON_ENTITY_HANDLE_DATA                                             \
-  SINCE (R_13) { dwg_free_common_entity_handle_data (obj); }
+#define COMMON_ENTITY_HANDLE_DATA
 #define SECTION_STRING_STREAM
 #define START_STRING_STREAM
 #define END_STRING_STREAM
@@ -265,13 +276,13 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
     int error = 0;                                                            \
     if (obj->tio.entity)                                                      \
       {                                                                       \
-        LOG_HANDLE ("Free entity " #token "\n")                               \
+        LOG_HANDLE ("Free entity " #token " [%d]\n", obj->index)              \
         if (obj->tio.entity->tio.token)                                       \
           error = dwg_free_##token##_private (dat, obj);                      \
                                                                               \
         dwg_free_common_entity_data (obj);                                    \
         dwg_free_eed (obj);                                                   \
-        if (obj->tio.object)                                                  \
+        if (obj->tio.entity)                                                  \
           {                                                                   \
             FREE_IF (obj->tio.entity->tio.token);                             \
             FREE_IF (obj->tio.entity);                                        \
@@ -294,7 +305,7 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
     _obj = ent = _ent->tio.token;
 
 #define DWG_ENTITY_END                                                        \
-  return 0;                                                                   \
+  return error;                                                               \
   }
 
 #define DWG_OBJECT(token)                                                     \
@@ -309,17 +320,8 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
     if (obj->tio.object)                                                      \
       {                                                                       \
         _obj = obj->tio.object->tio.token;                                    \
-        LOG_HANDLE ("Free object " #token " %p\n", obj)                       \
-        if (strcmp (#token, "UNKNOWN_OBJ")                                    \
-            && obj->supertype == DWG_SUPERTYPE_UNKNOWN)                       \
-          {                                                                   \
-            _obj = NULL;                                                      \
-            error = dwg_free_UNKNOWN_OBJ (dat, obj);                          \
-          }                                                                   \
-        else                                                                  \
-          {                                                                   \
-            error = dwg_free_##token##_private (dat, obj);                    \
-          }                                                                   \
+        LOG_HANDLE ("Free object " #token " [%d]\n", obj->index)              \
+        error = dwg_free_##token##_private (dat, obj);                        \
         dwg_free_eed (obj);                                                   \
         FREE_IF (_obj);                                                       \
         FREE_IF (obj->tio.object);                                            \
@@ -344,37 +346,18 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat,
 /* obj itself is allocated via dwg->object[], dxfname is klass->dxfname or
  * static */
 #define DWG_OBJECT_END                                                        \
-  return 0;                                                                   \
+  return error;                                                               \
   }
 
-static void
-dwg_free_common_entity_handle_data (Dwg_Object *obj)
-{
-
-  Dwg_Data *dwg = obj->parent;
-  Bit_Chain *dat = &pdat;
-  Dwg_Object_Entity *_obj;
-  BITCODE_BL vcount;
-  Dwg_Object_Entity *ent;
-  int error = 0;
-
-  ent = obj->tio.entity;
-  if (!ent)
-    return;
-  _obj = ent;
-
-#include "common_entity_handle_data.spec"
-}
-
 static void
 dwg_free_common_entity_data (Dwg_Object *obj)
 {
 
   Dwg_Data *dwg = obj->parent;
   Bit_Chain *dat = &pdat;
+  Bit_Chain *hdl_dat = &pdat;
   Dwg_Object_Entity *_obj;
   Dwg_Object_Entity *ent;
-  Bit_Chain *hdl_dat = NULL;
   BITCODE_BL vcount;
   int error = 0;
 
@@ -383,7 +366,14 @@ dwg_free_common_entity_data (Dwg_Object *obj)
     return;
   _obj = ent;
 
-#include "common_entity_data.spec"
+  FREE_IF (ent->preview);
+
+  // clang-format off
+  #include "common_entity_data.spec"
+  SINCE (R_13) {
+  #include "common_entity_handle_data.spec"
+  }
+  // clang-format on
 }
 
 static void
@@ -411,7 +401,7 @@ dwg_free_eed (Dwg_Object *obj)
   else
     {
       Dwg_Object_Object *_obj = obj->tio.object;
-      if (!_obj)
+      if (!_obj || !_obj->eed)
         return;
       for (i = 0; i < _obj->num_eed; i++)
         {
@@ -430,26 +420,34 @@ dwg_free_eed (Dwg_Object *obj)
 static int
 dwg_free_variable_type (Dwg_Data *restrict dwg, Dwg_Object *restrict obj)
 {
-  int i;
-  int is_entity;
+  const int i = obj->type - 500;
   Dwg_Class *klass;
   Bit_Chain *dat = &pdat;
 
-  i = obj->type - 500;
-  if (i < 0 || i > (int)dwg->num_classes)
+  if (i < 0 || i >= (int)dwg->num_classes)
     return DWG_ERR_INVALIDTYPE;
 
   klass = &dwg->dwg_class[i];
   if (!klass || !klass->dxfname)
     return DWG_ERR_INTERNALERROR;
-  // almost always false
-  is_entity = dwg_class_is_entity (klass);
 
-// global class dispatcher
-#include "classes.inc"
+  if (strNE (obj->dxfname, klass->dxfname))
+    {
+      LOG_ERROR ("Wrong %s.type %d for obj [%d]: != %s",  obj->dxfname, obj->type,
+                 obj->index, klass->dxfname);
+      return DWG_ERR_INVALIDTYPE;
+    }
 
-#undef WARN_UNHANDLED_CLASS
-#undef WARN_UNSTABLE_CLASS
+  // global class dispatcher:
+  // with indxf even DEBUGGING objects, such as TABLE are created.
+  // usually not written/encoded though.
+
+  // clang-format off
+  #include "classes.inc"
+
+  #undef WARN_UNHANDLED_CLASS
+  #undef WARN_UNSTABLE_CLASS
+  // clang-format on
 
   return DWG_ERR_UNHANDLEDCLASS;
 }
@@ -467,14 +465,12 @@ dwg_free_object (Dwg_Object *obj)
     {
       dwg = obj->parent;
       dat->version = dwg->header.version;
+      dat->from_version = dwg->header.from_version;
     }
   else
     return;
   if (obj->type == DWG_TYPE_FREED || obj->tio.object == NULL)
     return;
-  dat->from_version = dat->version;
-  if (obj->supertype == DWG_SUPERTYPE_UNKNOWN)
-    goto unhandled;
 
   switch (obj->type)
     {
@@ -718,45 +714,26 @@ dwg_free_object (Dwg_Object *obj)
       dwg_free_PROXY_OBJECT (dat, obj);
       break;
     default:
-      if (obj->type == obj->parent->layout_number)
+      if (obj->type == obj->parent->layout_type
+          && obj->fixedtype == DWG_TYPE_LAYOUT)
         {
           SINCE (R_13)
           {
             dwg_free_LAYOUT (dat, obj); // XXX avoid double-free, esp. in eed
           }
         }
-
       else if ((error = dwg_free_variable_type (obj->parent, obj))
                & DWG_ERR_UNHANDLEDCLASS)
         {
-          int is_entity;
-          int i;
-          Dwg_Class *klass;
-
-        unhandled:
-          i = obj->type - 500;
-          klass = NULL;
-
-          dwg = obj->parent;
-          if (dwg->dwg_class && i >= 0 && i < (int)dwg->num_classes)
-            {
-              klass = &dwg->dwg_class[i];
-              is_entity = klass ? dwg_class_is_entity (klass) : 0;
-            }
-          if (klass && !is_entity)
-            {
-              dwg_free_UNKNOWN_OBJ (dat, obj);
-            }
-          else if (klass)
-            {
+          if (obj->fixedtype == DWG_TYPE_UNKNOWN_ENT)
               dwg_free_UNKNOWN_ENT (dat, obj);
-            }
-          else // not a class
-            {
-              FREE_IF (obj->tio.unknown);
-            }
+          else if (obj->fixedtype == DWG_TYPE_UNKNOWN_OBJ)
+              dwg_free_UNKNOWN_OBJ (dat, obj);
         }
     }
+  /* With this importer the dxfname is dynamic, just the name is const */
+  if (dwg->opts & DWG_OPTS_INDXF)
+    FREE_IF (obj->dxfname);
   obj->type = DWG_TYPE_FREED;
 }
 
@@ -766,7 +743,22 @@ dwg_free_header_vars (Dwg_Data *dwg)
   Dwg_Header_Variables *_obj = &dwg->header_vars;
   Dwg_Object *obj = NULL;
   Bit_Chain *dat = &pdat;
-#include "header_variables.spec"
+  // clang-format off
+  #include "header_variables.spec"
+  // clang-format on
+  return 0;
+}
+
+static int
+dwg_free_summaryinfo (Dwg_Data *dwg)
+{
+  struct Dwg_SummaryInfo *_obj = &dwg->summaryinfo;
+  Dwg_Object *obj = NULL;
+  Bit_Chain *dat = &pdat;
+  BITCODE_RL rcount1, rcount2;
+  // clang-format off
+  #include "summaryinfo.spec"
+  // clang-format on
   return 0;
 }
 
@@ -777,9 +769,12 @@ dwg_free (Dwg_Data *dwg)
   if (dwg)
     {
       pdat.version = dwg->header.version;
-      pdat.from_version = dwg->header.version;
+      pdat.from_version = dwg->header.from_version;
       if (dwg->opts)
-        loglevel = dwg->opts & 0xf;
+        {
+          loglevel = dwg->opts & DWG_OPTS_LOGLEVEL;
+          pdat.opts = dwg->opts;
+        }
 #ifdef USE_TRACING
       /* Before starting, set the logging level, but only do so once.  */
       if (!env_var_checked_p)
@@ -799,8 +794,8 @@ dwg_free (Dwg_Data *dwg)
         }
       FREE_IF (dwg->header.section);
       dwg_free_header_vars (dwg);
-      if (dwg->picture.size && dwg->picture.chain)
-        FREE_IF (dwg->picture.chain);
+      dwg_free_summaryinfo (dwg);
+      FREE_IF (dwg->thumbnail.chain);
       if (dwg->header.section_infohdr.num_desc)
         {
           for (i = 0; i < dwg->header.section_infohdr.num_desc; ++i)
@@ -809,26 +804,32 @@ dwg_free (Dwg_Data *dwg)
         }
       for (i = 0; i < dwg->second_header.num_handlers; i++)
         FREE_IF (dwg->second_header.handlers[i].data);
+      // auxheader has no strings
       for (i = 0; i < dwg->num_objects; ++i)
         {
           if (dwg_obj_is_control (&dwg->object[i]))
             dwg_free_object (&dwg->object[i]);
         }
-      if (dwg->num_classes)
+      if (dwg->num_classes && dwg->dwg_class)
         {
           for (i = 0; i < dwg->num_classes; ++i)
             {
               FREE_IF (dwg->dwg_class[i].appname);
               FREE_IF (dwg->dwg_class[i].cppname);
               FREE_IF (dwg->dwg_class[i].dxfname);
-              if (dwg->header.version >= R_2007)
+              if (dwg->header.from_version >= R_2007)
                 FREE_IF (dwg->dwg_class[i].dxfname_u);
             }
           FREE_IF (dwg->dwg_class);
         }
-      for (i = 0; i < dwg->num_object_refs; ++i)
+      if (dwg->object_ref)
         {
-          FREE_IF (dwg->object_ref[i]);
+          LOG_HANDLE ("free %d global refs\n", dwg->num_object_refs)
+          for (i = 0; i < dwg->num_object_refs; ++i)
+            {
+              //LOG_HANDLE ("free ref %d\n", i)
+              FREE_IF (dwg->object_ref[i]);
+            }
         }
       FREE_IF (dwg->object_ref);
       FREE_IF (dwg->object);
