@@ -1,7 +1,7 @@
 /*****************************************************************************/
 /*  LibreDWG - free implementation of the DWG file format                    */
 /*                                                                           */
-/*  Copyright (C) 2018-2019 Free Software Foundation, Inc.                   */
+/*  Copyright (C) 2018-2020 Free Software Foundation, Inc.                   */
 /*                                                                           */
 /*  This library is free software, licensed under the terms of the GNU       */
 /*  General Public License as published by the Free Software Foundation,     */
@@ -103,8 +103,26 @@ static void _prefix (Bit_Chain *dat);
   dat->bit++;
 #define ENDSEC() ENDARRAY
 
+#undef FORMAT_H
+#undef FORMAT_RLL
+#undef FORMAT_BLL
 #undef FORMAT_RC
+#undef FORMAT_RSx
+#undef FORMAT_RLx
+#undef FORMAT_BLx
+#undef FORMAT_BLX
+#undef FORMAT_4BITS
+#define FORMAT_H "[%u, %lu]"
+#define FORMAT_RLL "%" PRIu64
+#define FORMAT_BLL "%" PRIu64
 #define FORMAT_RC "%d"
+#define FORMAT_RSx FORMAT_RS
+#define FORMAT_RLx FORMAT_RL
+#define FORMAT_BLx FORMAT_BL
+#define FORMAT_BLX FORMAT_BL
+#define FORMAT_4BITS FORMAT_RC
+#undef ARGS_H
+#define ARGS_H(hdl) hdl.code, hdl.value
 #define VALUE(value, type, dxf) fprintf (dat->fh, FORMAT_##type ",\n", value)
 #define VALUE_B(value, dxf) VALUE (value, B, dxf)
 #define VALUE_RC(value, dxf) VALUE (value, RC, dxf)
@@ -137,7 +155,6 @@ static void _prefix (Bit_Chain *dat);
     PREFIX fprintf (dat->fh, "\"" #nam "\": " FORMAT_##type ",\n",            \
                     _obj->o.nam);                                             \
   }
-
 #define FIELD_CAST(nam, type, cast, dxf) FIELD (nam, cast, dxf)
 #define FIELD_TRACE(nam, type)
 #define FIELD_G_TRACE(nam, type, dxf)
@@ -151,16 +168,18 @@ static void _prefix (Bit_Chain *dat);
         const int len = strlen (str);                                         \
         if (len < 4096 / 6)                                                   \
           {                                                                   \
-            char *_buf = alloca (6 * len + 1);                                \
+            const int _len = 6 * len + 1;                                     \
+            char *_buf = alloca (_len);                                       \
             fprintf (dat->fh, "\"" #nam "\": \"%s\",\n",                      \
-                     json_cquote (_buf, str));                                \
+                     json_cquote (_buf, str, _len));                          \
             freea (_buf);                                                     \
           }                                                                   \
         else                                                                  \
           {                                                                   \
-            char *_buf = malloc (6 * len + 1);                                \
+            const int _len = 6 * len + 1;                                     \
+            char *_buf = malloc (_len);                                       \
             fprintf (dat->fh, "\"" #nam "\": \"%s\",\n",                      \
-                     json_cquote (_buf, str));                                \
+                     json_cquote (_buf, str, _len));                          \
             free (_buf);                                                      \
           }                                                                   \
       }                                                                       \
@@ -198,42 +217,51 @@ static void _prefix (Bit_Chain *dat);
 #define VALUE_HANDLE(hdlptr, nam, handle_code, dxf)                           \
   if (hdlptr)                                                                 \
     {                                                                         \
-      fprintf (dat->fh, "\"%d.%d.%lu\",\n", hdlptr->handleref.code,           \
-               hdlptr->handleref.size, hdlptr->handleref.value);              \
+      fprintf (dat->fh, FORMAT_H ",\n", ARGS_H (hdlptr->handleref));          \
     }                                                                         \
   else                                                                        \
     {                                                                         \
-      fprintf (dat->fh, "\"0.0.0\",\n");                                      \
+      fprintf (dat->fh, "[0, 0],\n");                                         \
     }
-#define VALUE_H(hdl, dxf)                                                     \
-  fprintf (dat->fh, "\"%d.%d.%lu\",\n", (hdl).code, (hdl).size, (hdl).value)
+#define VALUE_H(hdl, dxf) fprintf (dat->fh, FORMAT_H ",\n", ARGS_H (hdl))
 #define FIELD_HANDLE(nam, handle_code, dxf)                                   \
-  PREFIX if (_obj->nam)                                                       \
   {                                                                           \
-    fprintf (dat->fh, "\"" #nam "\": \"%d.%d.%lu\",\n",                       \
-             _obj->nam->handleref.code, _obj->nam->handleref.size,            \
-             _obj->nam->handleref.value);                                     \
-  }                                                                           \
-  else { fprintf (dat->fh, "\"" #nam "\": \"0.0.0\",\n"); }
+    if (_obj->nam)                                                            \
+      {                                                                       \
+        PREFIX fprintf (dat->fh, "\"" #nam "\": " FORMAT_H ",\n",             \
+                        ARGS_H (_obj->nam->handleref));                       \
+      }                                                                       \
+    else                                                                      \
+      {                                                                       \
+        PREFIX fprintf (dat->fh, "\"" #nam "\": [0, 0],\n");                  \
+      }                                                                       \
+  }
 #define SUB_FIELD_HANDLE(o, nam, handle_code, dxf)                            \
-  PREFIX if (_obj->o.nam)                                                     \
   {                                                                           \
-    fprintf (dat->fh, "\"" #nam "\": \"%d.%d.%lu\",\n",                       \
-             _obj->o.nam->handleref.code, _obj->o.nam->handleref.size,        \
-             _obj->o.nam->handleref.value);                                   \
-  }                                                                           \
-  else { fprintf (dat->fh, "\"" #nam "\": \"0.0.0\",\n"); }
+    if (_obj->o.nam)                                                          \
+      {                                                                       \
+        PREFIX fprintf (dat->fh, "\"" #nam "\": " FORMAT_H ",\n",             \
+                        ARGS_H (_obj->o.nam->handleref));                     \
+      }                                                                       \
+    else                                                                      \
+      {                                                                       \
+        PREFIX fprintf (dat->fh, "\"" #nam "\": [0, 0],\n");                  \
+      }                                                                       \
+  }
 #define FIELD_DATAHANDLE(nam, code, dxf) FIELD_HANDLE (nam, code, dxf)
 #define FIELD_HANDLE_N(nam, vcount, handle_code, dxf)                         \
-  PREFIX if (_obj->nam)                                                       \
-  {                                                                           \
-    fprintf (dat->fh, "\"%d.%d.%lu\",\n", _obj->nam->handleref.code,          \
-             _obj->nam->handleref.size, _obj->nam->handleref.value);          \
-  }                                                                           \
-  else { fprintf (dat->fh, "\"0.0.0\",\n"); }
+  if (_obj->nam)                                                              \
+    {                                                                         \
+      PREFIX fprintf (dat->fh, FORMAT_H ",\n",                                \
+                      ARGS_H (_obj->nam->handleref));                         \
+    }                                                                         \
+  else                                                                        \
+    {                                                                         \
+      PREFIX fprintf (dat->fh, "[0, 0],\n");                                  \
+    }
 #define FIELD_BINARY(nam, size, dxf)                                          \
   {                                                                           \
-    long len = size;                                                          \
+    long len = (long)size;                                                    \
     KEY (nam);                                                                \
     fprintf (dat->fh, "\"");                                                  \
     if (_obj->nam)                                                            \
@@ -331,8 +359,10 @@ static void _prefix (Bit_Chain *dat);
 #define SUB_FIELD_BB(o, nam, dxf) SUB_FIELD (o, nam, BB, dxf)
 #define SUB_FIELD_3B(o, nam, dxf) SUB_FIELD (o, nam, 3B, dxf)
 #define SUB_FIELD_BS(o, nam, dxf) SUB_FIELD (o, nam, BS, dxf)
+#define SUB_FIELD_BSd(o, nam, dxf) SUB_FIELD (o, nam, BSd, dxf)
 #define SUB_FIELD_BL(o, nam, dxf) SUB_FIELD (o, nam, BL, dxf)
 #define SUB_FIELD_BLx(o, nam, dxf) SUB_FIELD (o, nam, BLx, dxf)
+#define SUB_FIELD_BLd(o, nam, dxf) SUB_FIELD (o, nam, BLd, dxf)
 #define SUB_FIELD_BD(o, nam, dxf) SUB_FIELD (o, nam, BD, dxf)
 #define SUB_FIELD_RC(o, nam, dxf) SUB_FIELD (o, nam, RC, dxf)
 #define SUB_FIELD_RS(o, nam, dxf) SUB_FIELD (o, nam, RS, dxf)
@@ -489,22 +519,6 @@ field_cmc (Bit_Chain *restrict dat, const char *restrict key,
 #define HANDLE_VECTOR(nam, sizefield, code, dxf)                              \
   HANDLE_VECTOR_N (nam, FIELD_VALUE (sizefield), code, dxf)
 
-#define REPEAT_CN(times, nam, type)                                           \
-  if (_obj->nam)                                                              \
-    {                                                                         \
-      KEY (nam);                                                              \
-      ARRAY;                                                                  \
-    }                                                                         \
-  if (_obj->nam)                                                              \
-    for (rcount1 = 0; rcount1 < (BITCODE_BL)times; rcount1++)
-#define REPEAT_N(times, nam, type)                                            \
-  if (_obj->nam)                                                              \
-    {                                                                         \
-      KEY (nam);                                                              \
-      ARRAY;                                                                  \
-    }                                                                         \
-  if (_obj->nam)                                                              \
-    for (rcount1 = 0; rcount1 < (BITCODE_BL)times; rcount1++)
 #define _REPEAT_N(times, nam, type, idx)                                      \
   if (_obj->nam)                                                              \
     {                                                                         \
@@ -513,8 +527,9 @@ field_cmc (Bit_Chain *restrict dat, const char *restrict key,
     }                                                                         \
   if (_obj->nam)                                                              \
     for (rcount##idx = 0; rcount##idx < (BITCODE_BL)times; rcount##idx++)
-#define _REPEAT_C(times, nam, type, idx)                                      \
-  _REPEAT_N (_obj->times, nam, type, idx)
+#define REPEAT_N(times, nam, type) _REPEAT_N(times, nam, type, 1)
+#define REPEAT_CN(times, nam, type) REPEAT_N(times, nam, type)
+#define _REPEAT_C(times, nam, type, idx) _REPEAT_N (_obj->times, nam, type, idx)
 #define _REPEAT(times, nam, type, idx) _REPEAT_N (_obj->times, nam, type, idx)
 #define REPEAT(times, nam, type) _REPEAT (times, nam, type, 1)
 #define REPEAT2(times, nam, type) _REPEAT (times, nam, type, 2)
@@ -524,6 +539,9 @@ field_cmc (Bit_Chain *restrict dat, const char *restrict key,
 #define REPEAT2_C(times, nam, type) _REPEAT_C (times, nam, type, 2)
 #define REPEAT3_C(times, nam, type) _REPEAT_C (times, nam, type, 3)
 #define REPEAT4_C(times, nam, type) _REPEAT_C (times, nam, type, 4)
+#define _REPEAT_CN(times, nam, type, idx) _REPEAT_N (times, nam, type, idx)
+#define _REPEAT_CNF(times, nam, type, idx) _REPEAT_N (times, nam, type, idx)
+#define _REPEAT_NF(times, nam, type, idx) _REPEAT_N (times, nam, type, idx)
 
 #undef REPEAT_BLOCK
 #define REPEAT_BLOCK                                                          \
@@ -548,40 +566,54 @@ field_cmc (Bit_Chain *restrict dat, const char *restrict key,
 #define FIELD_XDATA(nam, size)
 
 #define REACTORS(code)                                                        \
-  KEY (reactors);                                                             \
-  ARRAY;                                                                      \
-  if (obj->tio.object->reactors)                                              \
+  if (dat->version >= R_13 && obj->tio.object->num_reactors                   \
+      && obj->tio.object->reactors)                                           \
     {                                                                         \
+      KEY (reactors);                                                         \
+      ARRAY;                                                                  \
       for (vcount = 0; vcount < obj->tio.object->num_reactors; vcount++)      \
         {                                                                     \
           PREFIX VALUE_HANDLE (obj->tio.object->reactors[vcount], reactors,   \
                                code, 330);                                    \
         }                                                                     \
       NOCOMMA;                                                                \
-    }                                                                         \
-  ENDARRAY;
+      ENDARRAY;                                                               \
+    }
+#define ENT_REACTORS(code)                                                    \
+  if (dat->version >= R_13 && ent->num_reactors && ent->reactors)             \
+    {                                                                         \
+      KEY (reactors);                                                         \
+      ARRAY;                                                                  \
+      for (vcount = 0; vcount < ent->num_reactors; vcount++)                  \
+        {                                                                     \
+          PREFIX VALUE_HANDLE (ent->reactors[vcount], reactors, code, 330);   \
+        }                                                                     \
+      NOCOMMA;                                                                \
+      ENDARRAY;                                                               \
+    }
 
 #define XDICOBJHANDLE(code)                                                   \
-  SINCE (R_2004)                                                              \
-  {                                                                           \
-    if (!obj->tio.object->xdic_missing_flag)                                  \
-      {                                                                       \
-        KEY (xdicobjhandle);                                                  \
-        VALUE_HANDLE (obj->tio.object->xdicobjhandle, xdicobjhandle, code,    \
-                      -3);                                                    \
-      }                                                                       \
-  }                                                                           \
-  PRIOR_VERSIONS                                                              \
-  {                                                                           \
-    KEY (xdicobjhandle);                                                      \
-    VALUE_HANDLE (obj->tio.object->xdicobjhandle, xdicobjhandle, code, -3);   \
-  }
+  if ((dat->version < R_2004 || obj->tio.object->xdic_missing_flag != 0)      \
+      && (obj->tio.object->xdicobjhandle != NULL)                             \
+      && (obj->tio.object->xdicobjhandle->handleref.value != 0))              \
+    {                                                                         \
+      KEY (xdicobjhandle);                                                    \
+      VALUE_HANDLE (obj->tio.object->xdicobjhandle, xdicobjhandle, code, -3); \
+    }
+#define ENT_XDICOBJHANDLE(code)                                               \
+  if ((dat->version < R_2004 || (ent->xdic_missing_flag != 0))                \
+      && (ent->xdicobjhandle != NULL)                                         \
+      && (ent->xdicobjhandle->handleref.value != 0))                          \
+    {                                                                         \
+      KEY (xdicobjhandle);                                                    \
+      VALUE_HANDLE (ent->xdicobjhandle, xdicobjhandle, code, -3);             \
+    }
 
 #define COMMON_ENTITY_HANDLE_DATA
 #define SECTION_STRING_STREAM
 #define START_STRING_STREAM
 #define END_STRING_STREAM
-#define START_HANDLE_STREAM
+#define START_HANDLE_STREAM KEY (ownerhandle)
 
 static void
 _prefix (Bit_Chain *dat)
@@ -607,22 +639,16 @@ _prefix (Bit_Chain *dat)
     _ent = obj->tio.entity;                                                   \
     _obj = ent = _ent->tio.token;                                             \
     FIELD_TEXT (entity, #token);                                              \
-    if (strcmp (obj->dxfname, #token))                                        \
+    if (strNE (obj->dxfname, #token))                                         \
       FIELD_TEXT (dxfname, obj->dxfname);                                     \
     _FIELD (type, RL, 0);                                                     \
     KEY (handle);                                                             \
     VALUE_H (obj->handle, 5);                                                 \
     _FIELD (size, RL, 0);                                                     \
     _FIELD (bitsize, BL, 0);                                                  \
-    if (_ent->picture_exists)                                                 \
-      ENT_FIELD (picture_exists, B, 0);                                       \
-    SINCE (R_13)                                                              \
-    {                                                                         \
-      KEY (owner);                                                            \
-      VALUE_HANDLE (obj->parent->header_vars.BLOCK_RECORD_MSPACE, owner, 0,   \
-                    330);                                                     \
-      error |= json_common_entity_handle_data (dat, obj);                     \
-    }
+    if (_ent->preview_exists)                                                 \
+      ENT_FIELD (preview_exists, B, 0);                                       \
+    error |= json_common_entity_data (dat, obj);
 
 #define DWG_ENTITY_END                                                        \
   return 0;                                                                   \
@@ -640,7 +666,7 @@ _prefix (Bit_Chain *dat)
     LOG_INFO ("Object " #token ":\n")                                         \
     _obj = obj->tio.object->tio.token;                                        \
     FIELD_TEXT (object, #token);                                              \
-    if (strcmp (obj->dxfname, #token))                                        \
+    if (strNE (obj->dxfname, #token))                                         \
       FIELD_TEXT (dxfname, obj->dxfname);                                     \
     _FIELD (type, RL, 0);                                                     \
     KEY (handle);                                                             \
@@ -653,12 +679,23 @@ _prefix (Bit_Chain *dat)
   }
 
 static int
-json_common_entity_handle_data (Bit_Chain *restrict dat,
-                                Dwg_Object *restrict obj)
+json_common_entity_data (Bit_Chain *restrict dat,
+                         const Dwg_Object *restrict obj)
 {
-  (void)dat;
-  (void)obj;
-  return 0;
+  Dwg_Object_Entity *ent;
+  // Dwg_Data *dwg = obj->parent;
+  Dwg_Object_Entity *_obj;
+  int error = 0;
+  BITCODE_BL vcount = 0;
+  ent = obj->tio.entity;
+  _obj = ent;
+
+  // clang-format off
+  #include "common_entity_handle_data.spec"
+  #include "common_entity_data.spec"
+  // clang-format on
+
+  return error;
 }
 
 #include "dwg.spec"
@@ -759,34 +796,40 @@ wcquote (wchar_t *restrict dest, const wchar_t *restrict src)
 #endif /* HAVE_NATIVE_WCHAR2 */
 
 char *
-json_cquote (char *restrict dest, const char *restrict src)
+json_cquote (char *restrict dest, const char *restrict src, const int len)
 {
   unsigned char c;
   unsigned char *s = (unsigned char *)src;
+  const char* endp = dest + len;
   char *d = dest;
   while ((c = *s++))
     {
-      if (c == '"')
+      if (dest >= endp)
+        {
+          *dest = 0;
+          return d;
+        }
+      if (c == '"' && dest+1 < endp)
         {
           *dest++ = '\\';
           *dest++ = c;
         }
-      else if (c == '\\')
+      else if (c == '\\' && dest+1 < endp)
         {
           *dest++ = '\\';
           *dest++ = c;
         }
-      else if (c == '\n')
+      else if (c == '\n' && dest+1 < endp)
         {
           *dest++ = '\\';
           *dest++ = 'n';
         }
-      else if (c == '\r')
+      else if (c == '\r' && dest+1 < endp)
         {
           *dest++ = '\\';
           *dest++ = 'r';
         }
-      else if (c < 0x1f)
+      else if (c < 0x1f && dest+5 < endp)
         {
           *dest++ = '\\';
           *dest++ = 'u';
@@ -822,7 +865,9 @@ dwg_json_variable_type (Dwg_Data *restrict dwg, Bit_Chain *restrict dat,
   // almost always false
   is_entity = dwg_class_is_entity (klass);
 
-#include "classes.inc"
+  // clang-format off
+  #include "classes.inc"
+  // clang-format on
 
   return DWG_ERR_UNHANDLEDCLASS;
 }
@@ -991,7 +1036,7 @@ dwg_json_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
     case DWG_TYPE_LAYOUT:
       return dwg_json_LAYOUT (dat, obj);
     default:
-      if (obj->type == obj->parent->layout_number)
+      if (obj->type != 0 && obj->type == obj->parent->layout_type)
         {
           return dwg_json_LAYOUT (dat, obj);
         }
@@ -1023,8 +1068,8 @@ dwg_json_object (Bit_Chain *restrict dat, Dwg_Object *restrict obj)
               LOG_WARN ("Unknown object, skipping eed/reactors/xdic");
               SINCE (R_2000){
                 LOG_INFO ("Object bitsize: %u\n", obj->bitsize)
-              } LOG_INFO ("Object handle: %d.%d.%lX\n", obj->handle.code,
-                          obj->handle.size, obj->handle.value);
+              } LOG_INFO ("Object handle: " FORMAT_H "\n",
+                          ARGS_H (obj->handle));
             }
         }
     }
@@ -1042,7 +1087,9 @@ json_fileheader_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
   KEY (version);
   fprintf (dat->fh, "\"%s\",\n", version_codes[dwg->header.version]);
 
-#include "header.spec"
+  // clang-format off
+  #include "header.spec"
+  // clang-format on
 
   ENDRECORD ();
   return 0;
@@ -1062,7 +1109,9 @@ json_header_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
             : (dwg->header.version >= R_2007) ? "UTF-8" : "ANSI_1252";
 
   RECORD (HEADER); // single hash
-#include "header_variables.spec"
+                   // clang-format off
+  #include "header_variables.spec"
+  // clang-format on
   ENDRECORD ();
   return 0;
 }
@@ -1083,7 +1132,7 @@ json_classes_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
       FIELD_T (appname, 3);
       FIELD_BS (proxyflag, 90);
       FIELD_BL (num_instances, 91);
-      FIELD_B (wasazombie, 280);
+      FIELD_B (is_zombie, 280);
       FIELD_BS (item_class_id, 281);
       NOCOMMA;
       ENDHASH;
@@ -1131,9 +1180,9 @@ json_handles_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 }
 
 static int
-json_preview_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
+json_thumbnail_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 {
-  Bit_Chain *_obj = (Bit_Chain *)&dwg->picture;
+  Bit_Chain *_obj = (Bit_Chain *)&dwg->thumbnail;
   if (_obj->chain && _obj->size && _obj->size > 10)
     {
       KEY (THUMBNAILIMAGE);
@@ -1149,7 +1198,7 @@ json_preview_write (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 EXPORT int
 dwg_write_json (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 {
-  const int minimal = dwg->opts & 0x10;
+  const int minimal = dwg->opts & DWG_OPTS_MINIMAL;
   struct Dwg_Header *obj = &dwg->header;
 
   fprintf (dat->fh,
@@ -1180,7 +1229,7 @@ dwg_write_json (Bit_Chain *restrict dat, Dwg_Data *restrict dwg)
 
   if (!minimal && dat->version >= R_2000)
     {
-      if (json_preview_write (dat, dwg) >= DWG_ERR_CRITICAL)
+      if (json_thumbnail_write (dat, dwg) >= DWG_ERR_CRITICAL)
         goto fail;
     }
 
